from clock_tracking import logger
from clock_tracking import report
from clock_tracking import options
from clock_tracking import filters
from os.path import expanduser

class Clock():
    def __init__(self):
        self.reader = logger.ClockReader()
        self.writer = logger.ClockLogger()
        self.arg = options.ClockArguments()

        self.filters = []
        self.reports = []
        self.file = expanduser('~') + '/clock.txt'

    # Parses the arguments from the options.ClockArguments() and 
    # fills in the self.file, self.filters and self.reports
    def parse_arguments(self):
        self.arg.parse()
        if self.arg.options.file is not None and self.arg.options.file != '':
            self.file = self.arg.options.file
        self.filters = self.arg.get_filters()
        if self.arg.options.command != 'show':
            return

        if self.arg.options.details:
            self.reports.append(report.DetailsReport())
        elif self.arg.options.timeline and self.arg.options.today:
            self.reports.append(report.DayTimelineReport())
        elif len(self.arg.arguments) > 0 and self.arg.arguments[0] == 'tags':
            self.arg.options.categories = False
            self.reports.append(report.TagsReport())
        elif self.arg.options.categories:
            self.reports.append(report.CategoriesReport(len(self.arg.arguments)))
        if self.arg.options.details or self.arg.options.timeline or self.arg.options.categories:
            self.reports.append(report.TotalTimeReport(self.arg.get_target_time()))

    # Filters issues according to self.filters and shows reports in self.reports
    def show(self):
        self.reader.read_file(self.file)
        collection = self.reader.parse()
        for f in self.filters:
            collection = f.apply_to(collection)
        for r in self.reports:
            r.print_report(collection)

    # Add new entry at given time (at) with given description (description)
    def add(self, at, description):
        self.writer.read_file(self.file)
        self.writer.add(at, description)
        self.writer.write_file(self.file)

    # Edits current entry by replacing its description by given description
    def edit(self, description):
        self.writer.read_file(self.file)
        self.writer.edit_current(description)
        self.writer.write_file(self.file)

    def restart(self, at):
        self.writer.read_file(self.file)
        desc = self.writer.get_last()
        self.writer.add(at, desc)
        self.writer.write_file(self.file)

    # Shows current issue report
    @staticmethod
    def report_current(file):
        c = Clock()
        c.file = file
        c.filters.append(filters.LastFilter(1))
        c.reports.append(report.DetailsReport())
        c.show()

    # Single static method to run script according to command line arguments
    @staticmethod
    def run():
        clock = Clock()
        clock.parse_arguments()
        options = clock.arg.options

        if options.command == 'show':
            clock.show()
        elif options.command == 'add':
            clock.add(options.at, ' '.join(clock.arg.arguments))
        elif options.command == 'edit':
            clock.edit(' '.join(clock.arg.arguments))
        elif options.command == 'stop':
            clock.add(options.at, '[Stop]')
        elif options.command == 'restart':
            clock.restart(options.at)

        if options.command in ('add', 'edit', 'stop', 'restart'):
            Clock.report_current(options.file)



import datetime

class Period:
    def __init__(self, start):
        self.start = start
        self.end = datetime.datetime(1900,1,1)

class Task:
    def __init__(self, description, start):
        self.description = ''
        self.periods = []
        self.tags = []
        self.ids = []
        self.parse(description)
        self.periods.append(Period(start))
    
    def parse(self, description):
        description = description.replace('\r', '').replace('\n', '')
        words = description.split(' ')
        desc = ''
        for word in words:
            if len(word) == 0:
                continue
            if word[0] == '+' and len(word) > 1:
                self.tags.append(word.strip())
            elif word[0] == '.' and len(word) > 1:
                self.ids.append(word.strip())
            else:
                desc += word + ' '
        self.description = desc.strip()
        if len(self.ids) > 0:
            self.tags.insert(0, '+jira')

    def is_stop(self):
        return '[Stop]' in self.description

    def finish(self, end):
        if len(self.periods) > 0:
            self.periods[-1].end = end

    def discard_last(self):
        if len(self.periods) == 0:
            return
        self.periods.pop()

    def is_matching(self, parent_tags):
        for i in range(0, len(parent_tags)):
            if len(self.tags) <= i:
                return False
            if self.tags[i] != parent_tags[i]:
                return False
        return True

    def duration_total(self):
        total_time = datetime.timedelta()
        for period in self.periods:
            total_time += period.end - period.start
        return total_time

    def equals(self, task):
        if self.description != task.description:
            return False
        for tag in self.tags:
            if not tag in task.tags:
                return False
        return True
    
    def copy_empty(self):
        t = Task(self.description, datetime.datetime.now())
        t.tags = self.tags
        t.ids = self.ids
        t.periods = []
        return t

class TaskCollection:
    def __init__(self):
        self.tasks = []

    def exists(self, task):
        for t in self.tasks:
            if t.equals(task):
                return True
        return False

    def get_task(self, task):
        for t in self.tasks:
            if t.equals(task):
                return t
        return None

    def add_task(self, task):
        if task.is_stop(): # Do not add [Stop] tasks
            return task
        # If task already exists, add its periods
        if self.exists(task): 
            t = self.get_task(task)
            for p in task.periods:
                t.periods.append(p)
            return t

        self.tasks.append(task)
        return task

    def filter(self, filter):
        c = TaskCollection()
        c.tasks = [filter.get_task(task) for task in self.tasks if filter.is_valid(task)]
        return c


import unittest
import datetime
from clock_tracking.logger import ClockLogger, ClockReader

class TestClockLoggerDateExists(unittest.TestCase):
    def test__empty__returns_false(self):
        logger = ClockLogger()
        self.assertFalse(logger.date_exists())

    def test__single_not_today__returns_false(self):
        logger = ClockLogger()
        logger.lines = ['[2022-01-01]\n']
        self.assertFalse(logger.date_exists())

    def test__single_today__returns_true(self):
        logger = ClockLogger()
        logger.lines = ['[' + datetime.datetime.today().strftime('%Y-%m-%d')+ ']\n']
        self.assertTrue(logger.date_exists())

    def test_dual_with_today__returns_true(self):
        logger = ClockLogger()
        logger.lines = ['[2022-01-01]\n', '[' + datetime.datetime.today().strftime('%Y-%m-%d')+ ']\n']
        self.assertTrue(logger.date_exists())

class TestClockLoggerEditCurrent(unittest.TestCase):
    def test__empty__does_nothing(self):
        logger = ClockLogger()
        logger.edit_current('Test')
        self.assertTrue(len(logger.lines) == 0)

    def test__one_line__edits_line(self):
        logger = ClockLogger()
        logger.lines = ['09:00 Test\n']
        logger.edit_current('Test2')
        self.assertEqual(len(logger.lines), 1)
        self.assertEqual(logger.lines[0], '09:00 Test2\n')

class TestClockLoggerAdd(unittest.TestCase):
    def test__with_today__adds_one_line(self):
        logger = ClockLogger()
        logger.lines = ['[' + datetime.datetime.today().strftime('%Y-%m-%d') + ']\n']
        logger.add('09:00', 'Test')
        self.assertTrue(len(logger.lines) == 2)
        self.assertEqual(logger.lines[1], '09:00 Test\n')
    
    def test__empty__adds_today_with_line(self):
        logger = ClockLogger()
        logger.add('09:00', 'Test')
        self.assertTrue(len(logger.lines) == 2)
        self.assertEqual(logger.lines[1], '09:00 Test\n')

class TestClockReaderParse(unittest.TestCase):
    def test__date_only__returns_empty(self):
        reader = ClockReader()
        reader.lines = ['[2022-01-01]']
        c = reader.parse()
        self.assertTrue(len(c.tasks) == 0)

    def test__date_and_task__returns_one_task(self):
        reader = ClockReader()
        reader.lines = ['[2022-01-01]', '09:00 task1']
        c = reader.parse()
        self.assertTrue(len(c.tasks) == 1)

    def test__task_without_date__returns_task_with_default_date(self):
        reader = ClockReader()
        reader.lines = ['09:00 task1']
        c = reader.parse()
        self.assertTrue(len(c.tasks) == 1)
        self.assertTrue(len(c.tasks[0].periods) == 1)
        self.assertTrue(c.tasks[0].periods[0].start.date().year == 1900)

    def test__missing_stop__removes_last_task_of_day(self):
        reader = ClockReader()
        reader.lines = ['[2022-01-01]', '09:00 task without stop', '[2022-01-02]']
        c = reader.parse()
        self.assertTrue(len(c.tasks) == 1)
        self.assertTrue(len(c.tasks[0].periods) == 0)

    def test__task_with_stop__finishes_task(self):
        reader = ClockReader()
        reader.lines = ['[2022-01-01]', '09:00 task with stop', '09:10 [Stop]']
        c = reader.parse()
        self.assertTrue(len(c.tasks) == 1)
        self.assertTrue(len(c.tasks[0].periods) == 1)
        self.assertTrue(c.tasks[0].duration_total().seconds == 600)

    def test__same_tasks__adds_periods(self):
        reader = ClockReader()
        reader.lines = ['[2022-01-01]', '09:00 task', '10:00 task2', '11:00 task', '12:00 [Stop]']
        c = reader.parse()
        self.assertTrue(len(c.tasks) == 2)
        self.assertTrue(len(c.tasks[0].periods) == 2)
        self.assertTrue(c.tasks[0].duration_total().seconds == 7200)

    def test__empty_day__does_not_add_task(self):
        reader = ClockReader()
        reader.lines = ['[2022-01-01]', '09:00 task', '10:00 [Stop]', '[2022-01-02]', '[2022-01-03]', '09:00 task1']
        c = reader.parse()
        self.assertTrue(len(c.tasks) == 2)



import datetime

class DateTimeUtils:
    @staticmethod
    def get_date(line):
        return datetime.datetime.strptime(line[1:11], '%Y-%m-%d')

    @staticmethod
    def is_date(line):
        if line[0] == '[':
            try:
                DateTimeUtils.get_date(line)
                return True
            except ValueError:
                return False

    @staticmethod
    def show_date(date):
        return date.strftime('%Y-%m-%d')

    @staticmethod
    def show_time(date):
        return date.strftime('%H:%M')

    @staticmethod
    def parse_time(date, time):
        day_time = datetime.datetime.strptime(time, '%H:%M')
        return datetime.datetime(date.year, date.month, date.day, day_time.hour, day_time.minute, 0)

    @staticmethod
    def parse_duration(duration):
        s = duration.split(':')
        if len(s) == 0:
            return datetime.timedelta(0)
        if len(s) == 1:
            return datetime.timedelta(0, int(s) * 3600)
        if len(s) == 2:
            return datetime.timedelta(0, int(s[0]) * 3600 + int(s[1]) * 60)
        if len(s) == 3:
            return datetime.timedelta(0, int(s[0]) * 3600 + int(s[1]) * 60 + int(s[2]))

        return datetime.timedelta(0)

    @staticmethod
    def get_seconds(timedelta):
        return timedelta.days * 24 * 3600 + timedelta.seconds

    @staticmethod
    def show_timedelta(delta):
        total_seconds = DateTimeUtils.get_seconds(delta)
        hours = int(total_seconds / 3600)
        minutes = int((total_seconds - hours * 3600) / 60)
        return str(hours).zfill(2) + 'h ' + str(minutes).zfill(2) + 'm'

    



import unittest
import datetime
from clock_tracking.task import Task, Period, TaskCollection

class TestTaskParse(unittest.TestCase):
    def test__single_description__parses_correctly(self):
        t = Task('this is a test', datetime.datetime.now())
        self.assertTrue(t.description == 'this is a test')
        self.assertTrue(len(t.tags) == 0)
        self.assertTrue(len(t.periods) == 1)

    def test__description_with_one_tag__parse_one_tag(self):
        t = Task("this is a description with one +tag", datetime.datetime.now())
        self.assertTrue(len(t.tags) == 1)
        self.assertTrue(t.tags[0] == '+tag')

    def test__description_with_two_tags__parses_two_tags(self):
        t = Task("this is a description with two +tag +tag2", datetime.datetime.now())
        self.assertTrue(len(t.tags) == 2)
        self.assertTrue(t.tags[1] == '+tag2')

    def test__with_id__parses_id(self):
        t = Task("test with id .4537", datetime.datetime.now())
        self.assertTrue(len(t.ids) == 1)
        self.assertTrue(t.ids[0] == '.4537')

    def test__with_id__adds_jira_tag(self):
        t = Task("test with id .4537", datetime.datetime.now())
        self.assertTrue(len(t.tags) == 1)
        self.assertTrue(t.tags[0] == '+jira')

    def test__with_id__removes_id_from_description(self):
        t = Task("test with id .4537", datetime.datetime.now())
        self.assertTrue(t.description == 'test with id')

    def test__with_new_line__removes_new_line(self):
        t = Task("test with new line\n", datetime.datetime.now())
        self.assertTrue(t.description == 'test with new line')

    def test__with_cariage_return__removes_new_line(self):
        t = Task("test with new line\r", datetime.datetime.now())
        self.assertTrue(t.description == 'test with new line')

class TestTaskIsStop(unittest.TestCase):
    def test__when_stop__returns_true(self):
        t = Task('[Stop]', datetime.datetime.now())
        self.assertTrue(t.is_stop())

    def test__with_spaces__returns_true(self):
        t = Task(' [Stop]  ', datetime.datetime.now())
        self.assertTrue(t.is_stop())

    def test__not_stop__returns_false(self):
        t = Task(' [Stop', datetime.datetime.now())
        self.assertTrue(not t.is_stop())

class TestTaskFinish(unittest.TestCase):
    def test__no_periods__does_not_throw(self):
        t = Task('test', datetime.datetime.now())
        t.periods = []
        t.finish(datetime.datetime.now())

    def test__one_task__finishes_task(self):
        t = Task('test', datetime.datetime.now())
        t.finish(datetime.datetime.now())
        self.assertTrue(len(t.periods) == 1)
        self.assertTrue(t.periods[0].end.year != 1900)

class TestTaskDiscardLast(unittest.TestCase):
    def test__no_periods__does_not_throw(self):
        t = Task('test', datetime.datetime.now())
        t.periods = []
        t.discard_last()

    def test__one_period__clears_periods(self):
        t = Task('test', datetime.datetime.now())
        t.discard_last()
        self.assertTrue(len(t.periods) == 0)

class TestTaskIsMatching(unittest.TestCase):
    def test__no_tag__returns_true(self):
        t = Task('test', datetime.datetime.now())
        self.assertTrue(t.is_matching([]))

    def test__one_tag__returns_true(self):
        t = Task('test +cat', datetime.datetime.now())
        self.assertTrue(t.is_matching(['+cat']))

    def test__one_tag_different__returns_false(self):
        t = Task('test +cat', datetime.datetime.now())
        self.assertTrue(not t.is_matching(['+test']))

    def test__two_tags__returns_true(self):
        t = Task('test +cat1 +cat2', datetime.datetime.now())
        self.assertTrue(t.is_matching(['+cat1', '+cat2']))

    def test__two_tags_wrong_order__returns_false(self):
        t = Task('test +cat2 +cat1', datetime.datetime.now())
        self.assertTrue(not t.is_matching(['+cat1', '+cat2']))

    def test__one_tag_testing_two__returns_false(self):
        t = Task('test +cat', datetime.datetime.now())
        self.assertTrue(not t.is_matching(['+cat', '+cat2']))

class TestTaskDurationTotal(unittest.TestCase):
    def test__empty__returns_zero(self):
        t = Task('test', datetime.datetime.now())
        t.periods = []
        self.assertTrue(t.duration_total().seconds == 0)

    def test__one__returns_one(self):
        t = Task('test', datetime.datetime(2022,1,1,10,0,0))
        t.finish(datetime.datetime(2022,1,1,11,0,0))
        self.assertTrue(t.duration_total().seconds == 3600)

    def test__two__returns_sum(self):
        t = Task('test', datetime.datetime(2022, 1, 1, 10, 0, 0))
        t.finish(datetime.datetime(2022, 1, 1, 10, 30, 0))
        p = Period(datetime.datetime(2022,1,1,12,0,0))
        p.end = datetime.datetime(2022, 1, 1, 12, 30, 0)
        t.periods.append(p)
        self.assertTrue(t.duration_total().seconds == 3600)

class TestTaskEquals(unittest.TestCase):
    def test__different_description__returns_false(self):
        t = Task('test1', datetime.datetime.now())
        t2 = Task('test2', datetime.datetime.now())
        self.assertTrue(not t.equals(t2))

    def test__different_tags__returns_false(self):
        t = Task('test +tag', datetime.datetime.now())
        t2 = Task('test', datetime.datetime.now())
        self.assertTrue(not t.equals(t2))

    def test__are_equal__returns_true(self):
        t = Task('test +tag', datetime.datetime.now())
        t2 = Task('test +tag', datetime.datetime.now())
        self.assertTrue(t.equals(t2))

class TestTaskCopyEmpty(unittest.TestCase):
    def test__non_empty_task__returns_empty_task(self):
        t = Task('test', datetime.datetime.now())
        self.assertTrue(len(t.periods) != 0)
        t2 = t.copy_empty()
        self.assertTrue(len(t2.periods) == 0)

class TestTaskCollectionExists(unittest.TestCase):
    def test__existing_task__return_true(self):
        c = TaskCollection()
        task = Task('test', datetime.datetime.now())
        c.tasks.append(task)
        self.assertTrue(c.exists(task))

    def test__non_existing_task__returns_false(self):
        c = TaskCollection()
        task = Task('test', datetime.datetime.now())
        self.assertTrue(not c.exists(task))

class TestTaskCollectionGetTask(unittest.TestCase):
    def test__non_existing_task__returns_none(self):
        c = TaskCollection()
        task = Task('test', datetime.datetime.now())
        self.assertTrue(c.get_task(task) is None)

    def test__existing_task__returns_task(self):
        c = TaskCollection()
        task = Task('test', datetime.datetime.now())
        c.tasks.append(task)
        self.assertTrue(c.get_task(task).equals(task))
    
class TestTaskCollectionAddTask(unittest.TestCase):
    def test__stop__does_not_add_task(self):
        c = TaskCollection()
        task = Task('[Stop]', datetime.datetime.now())
        c.add_task(task)
        self.assertTrue(len(c.tasks) == 0)

    def test__non_existing_task__adds_task(self):
        c = TaskCollection()
        task = Task('test', datetime.datetime.now())
        c.add_task(task)
        self.assertTrue(len(c.tasks) == 1)

    def test__existing_task__adds_periods(self):
        c = TaskCollection()
        task = Task('test', datetime.datetime(2022, 1, 1, 9))
        task.finish(datetime.datetime(2022, 1, 1, 10))
        c.add_task(task)
        task2 = Task('test', datetime.datetime(2022, 1, 1, 11))
        c.add_task(task2)
        self.assertTrue(len(c.tasks) == 1)
        self.assertTrue(len(c.tasks[0].periods) == 2)

    def test__existing_task__returns_existing_task(self):
        c = TaskCollection()
        task = Task('test', datetime.datetime(2022, 1, 1, 9))
        task.finish(datetime.datetime(2022, 1, 1, 10))
        c.add_task(task)
        task2 = Task('test', datetime.datetime(2022, 1, 1, 11))
        result = c.add_task(task2)
        self.assertTrue(result == task)




from clock_tracking.task import Task
from clock_tracking.task import TaskCollection

class TaskFilterBase:
    def is_valid(self, task):
        return True
    def get_task(self, task):
        return task
    def apply_to(self, collection):
        c = TaskCollection()
        c.tasks = [t for t in collection.tasks if self.is_valid(t)]
        return c

class TagFilter(TaskFilterBase):
    def __init__(self, args):
        self.parent_tags = []
        for arg in args:
            if '+' in arg:
                self.parent_tags.append(arg)

    def is_valid(self, task):
        return task.is_matching(self.parent_tags)

class IdFilter(TaskFilterBase):
    def __init__(self, args):
        self.ids = []
        for arg in args:
            if '.' in arg:
                self.ids.append(arg)
    
    def is_valid(self, task):
        if len(self.ids) == 0:
            return True

        for id in self.ids:
            if id in task.ids:
                return True
        return False

class DateFilter(TaskFilterBase):
    def __init__(self, date):
        self.date = date

    def is_valid(self, task):
        if len(task.periods) == 0:
            return False
        for p in task.periods:
            if p.start.date() == self.date.date():
                return True
        return False

    def get_task(self, task):
        t = task.copy_empty()
        for p in task.periods:
            if p.start.date() == self.date.date():
                t.periods.append(p)

        return t

    def apply_to(self, collection):
        c = TaskCollection()
        c.tasks = [self.get_task(t) for t in collection.tasks if self.is_valid(t)]
        return c

class PeriodFilter(TaskFilterBase):
    def __init__(self, date_start, date_end):
        self.date_start = date_start.date()
        self.date_end = date_end.date()

    def is_valid(self, task):
        if len(task.periods) == 0:
            return False
        for p in task.periods:
            if p.start.date() >= self.date_start and p.start.date() <= self.date_end: # TODO we should probably also check the p.end.date() but it's supposed to be the same date in current architecture
                return True
        return False
    
    def get_task(self, task):
        t = task.copy_empty()
        for p in task.periods:
            if p.start.date() >= self.date_start and p.start.date() <= self.date_end:
                t.periods.append(p)

        return t

    def apply_to(self, collection):
        c = TaskCollection()
        c.tasks = [self.get_task(t) for t in collection.tasks if self.is_valid(t)]
        return c

class LastFilter(TaskFilterBase):
    def __init__(self, count):
        self.count = count

    def apply_to(self, collection):
        c = TaskCollection()
        periods = []
        for task in collection.tasks:
            for p in task.periods:
                periods.append([p, task])

        ordered_periods = sorted(periods, key=lambda t: t[0].start, reverse=True)
        for i in range(self.count):
            if i < len(ordered_periods):
                t = ordered_periods[i][1].copy_empty()
                t.periods.append(ordered_periods[i][0])
                c.tasks.append(t)

        return c




import configparser
from clock_tracking.datetimeutils import DateTimeUtils

class ClockConfig:
    def __init__(self, file):
        self.config = configparser.ConfigParser()
        self.config.read_file(file)

    def target(self):
        return DateTimeUtils.parse_duration(self.config.get('Target', 'total_target', fallback='00:00'))
    def target_per_day(self):
        return DateTimeUtils.parse_duration(self.config.get('Target', 'target_per_day', fallback='00:00'))
    def file(self):
        return self.config.get('File', 'file', fallback='')



from datetime import datetime
from datetime import timedelta
import argparse
import sys
from os.path import expanduser
from clock_tracking.datetimeutils import DateTimeUtils
from clock_tracking import filters

class TargetTime:
    def __init__(self, target, is_per_day):
        self.target = target
        self.is_per_day = is_per_day

class ClockCommands:
    list = ['add', 'edit', 'stop', 'show', 'restart']

class ClockArguments:
    def __init__(self):
        self.input_args = sys.argv[1:]
        self.options = None
        self.arguments = []
    
    def parse(self):
        parser = argparse.ArgumentParser(description='Helps managing time tracking from the command-line')
        parser.add_argument('command', default='add', help='Command (add, stop, edit, show). add: add a new entry. stop: stop current entry. edit: edit current entry\'s description. show: show reports and statistics.')
        settings_group = parser.add_argument_group('settings')
        settings_group.add_argument('-f', '--file', type=str, help='Speficy the file to store time entries. Default is ~/clock.txt')
        settings_group.add_argument('--target', type=str, metavar='HH:MM', help='<show> Sets expected target time (format HH:MM) and computes the difference with actual times in the reports')
        settings_group.add_argument('--target-per-day', type=str, metavar='HH:MM', help='<show> Sets expected target time per day (format HH:MM) and computes the difference with actual times in the reports')
        add_group = parser.add_argument_group('add')
        add_group.add_argument('-a', '--at', type=str, metavar='HH:MM', default=datetime.today().strftime('%H:%M'), help='<add> Specify a time (format HH:MM) of a new entry')
        # Filters
        filter_group = parser.add_argument_group('filters')
        filter_group.add_argument('-t', '--today', action='store_true', help='<show> Show only entries from today')
        filter_group.add_argument('-w', '--week', action='store_true', help='<show> Show only entries from the current week')
        filter_group.add_argument('-s', '--from', type=str, metavar='YYYY-mm-dd', dest='from_', help='<show> Include entries with start date later or equal to given date (format YYYY-mm-dd)')
        filter_group.add_argument('-e', '--to', type=str, metavar='YYYY-mm-dd', help='<show> Include entries with start date earlier or equal to given date (format YYYY-mm-dd)')
        filter_group.add_argument('-l', '--last', type=int, metavar='n', help='<show> Show only the last n entries', default=0)
        # Reports
        report_group = parser.add_argument_group('reports')
        report_group.add_argument('-d', '--details', action='store_true', help='<show> Shows detailed report')
        report_group.add_argument('--categories', action='store_true', help='<show> Shows categories report (default)', default=True)
        report_group.add_argument('--timeline', action='store_true', help='<show> Shows issues on a timeline (only when --today is specified)')

        if len(self.input_args) == 0:
            self.input_args = ['show']

        opt, args = parser.parse_known_args(self.input_args)
        if opt.command not in ClockCommands.list:
            args.insert(0, opt.command)
            opt.command = 'add'

        if opt.file is None or opt.file == '':
            opt.file = expanduser('~') + '/clock.txt'

        self.options = opt
        self.arguments = args

    def get_target_time(self):
        if self.options.target:
            return TargetTime(DateTimeUtils.parse_duration(self.options.target), False)
        elif self.options.target_per_day:
            return TargetTime(DateTimeUtils.parse_duration(self.options.target_per_day), True)
        return TargetTime(timedelta(0), False)

    def get_filters(self):
        task_filters = []
        if self.options.today:
            task_filters.append(filters.DateFilter(datetime.today()))
        elif self.options.week:
            today = datetime.today()
            monday = today - timedelta(days=today.weekday())
            sunday = monday + timedelta(days=6)
            task_filters.append(filters.PeriodFilter(monday, sunday))
        elif self.options.from_ or self.options.to:
            from_date = datetime(1900, 1, 1)
            to_date = datetime(2100, 1, 1)
            if self.options.from_:
                from_date = datetime.strptime(self.options.from_, '%Y-%m-%d')
            if self.options.to:
                to_date = datetime.strptime(self.options.to, '%Y-%m-%d')
            task_filters.append(filters.PeriodFilter(from_date, to_date))
        if self.options.last > 0:
            task_filters.append(filters.LastFilter(self.options.last))
        if len(self.arguments) > 0:
            task_filters.append(filters.TagFilter(self.arguments))
            task_filters.append(filters.IdFilter(self.arguments))
        return task_filters

   


import datetime
from xmlrpc.client import DateTime
from clock_tracking.datetimeutils import DateTimeUtils
from clock_tracking.colors import TerminalColors
import os

class PrintHelpers:
    color_index = -1

    @staticmethod
    def get_graph(time, totaltime, char_width_total):
        char_width = DateTimeUtils.get_seconds(time) * char_width_total / DateTimeUtils.get_seconds(totaltime)
        result = [u'\u2588' for i in range(int(char_width))]
        result.extend([' ' for i in range(int(char_width), char_width_total)])
        return ''.join(result)

    @staticmethod
    def get_color():
        PrintHelpers.color_index = PrintHelpers.color_index + 1
        if PrintHelpers.color_index > 2:
            PrintHelpers.color_index = 0

        if PrintHelpers.color_index == 0:
            return TerminalColors.BLUE
        elif PrintHelpers.color_index == 1:
            return TerminalColors.CYAN
        return TerminalColors.GREEN

    @staticmethod
    def max_width(string, width):
        if len(string) < width:
            return string.ljust(width)

        return (string[0:(width-3)] + '...').ljust(width)

    @staticmethod
    def colorize(string, color):
        if not TerminalColors.enabled:
            return string
        return color + string + TerminalColors.END


class TaskReportBase(object):
    def __init__(self):
        pass

    def print_report(self, collection):
        print('')

# Prints periods in chronological order
class DetailsReport(TaskReportBase):
    def __init__(self):
        super(DetailsReport, self).__init__()

    def print_report(self, collection):
        periods = []
        for task in collection.tasks:
            for p in task.periods:
                periods.append([p, task])

        periods = sorted(periods, key=lambda t: t[0].start)
        total_width = os.get_terminal_size().columns
        remaining_width = total_width - (10 + 12 + 6 + 6 + 20 + 10)
        if remaining_width < 0:
            print("Terminal is too small to show details.")
            exit

        print(str('Duration').ljust(10) + 'Date'.ljust(12) + 'Start'.ljust(6) + 'Stop'.ljust(6) + 'Tags'.ljust(20) + 'IDs'.ljust(10) + 'Name'.ljust(remaining_width))
        for kv in periods:
            p = kv[0]
            print(PrintHelpers.colorize(DateTimeUtils.show_timedelta(p.end - p.start).ljust(10), TerminalColors.GREEN) \
                + DateTimeUtils.show_date(p.start).ljust(12) \
                + DateTimeUtils.show_time(p.start).ljust(6) \
                + DateTimeUtils.show_time(p.end).ljust(6) \
                + PrintHelpers.colorize(PrintHelpers.max_width(','.join(kv[1].tags), 20), TerminalColors.BLUE) \
                + PrintHelpers.colorize(PrintHelpers.max_width(','.join(kv[1].ids), 10), TerminalColors.BOLD) \
                + PrintHelpers.max_width(kv[1].description, remaining_width))

# Print total time for given collection of entries
class TotalTimeReport(TaskReportBase):
    def __init__(self, target_time):
        super(TotalTimeReport, self).__init__()
        self.target_time = target_time

    # Gets targetted times (both total and per day)
    def get_targets(self, total_duration, days):
        if self.target_time.target == datetime.timedelta(0):
            return '', ''
        target_total = self.target_time.target
        if self.target_time.is_per_day:
            target_total = target_total * len(days)
        
        total_difference = total_duration - target_total
        return ' (' + DateTimeUtils.show_timedelta(total_difference) + ')', ' (' + DateTimeUtils.show_timedelta(total_difference / len(days)) + ')'

    def print_report(self, collection):
        total_duration = datetime.timedelta(0)
        days = {}
        for task in collection.tasks:
            total_duration += task.duration_total()
            if len(task.periods) > 0 and str(task.periods[0].start.date()) not in days:
                days[str(task.periods[0].start.date())] = 1

        if len(days) == 0:
            return

        target_total, target_per_day = self.get_targets(total_duration, days)
        print('')
        print(PrintHelpers.colorize(DateTimeUtils.show_timedelta(total_duration), TerminalColors.GREEN) + target_total + ' Total')
        if len(days) > 1:
            print(PrintHelpers.colorize(DateTimeUtils.show_timedelta(total_duration / len(days)), TerminalColors.GREEN) + target_per_day + ' Per day')

# Print graphical report by categories
class CategoriesReport(TaskReportBase):
    def __init__(self, filter_level):
        super(CategoriesReport, self).__init__()
        self.filter_level = filter_level

    def get_categories(self, collection, level):
        tags = {}
        if len(collection.tasks) == 0:
            return tags

        for task in collection.tasks:
            tag_name = task.description
            if len(task.tags) > level:
                tag_name = task.tags[level]
            if tag_name in tags:
                tags[tag_name] += task.duration_total()
            else:
                tags[tag_name] = task.duration_total()
        
        return sorted(tags.items(), key=lambda kv: kv[1], reverse=True)

    def get_graph_length(self, max_tag_length):
        graph_length = os.get_terminal_size().columns - max_tag_length - 10
        if graph_length > 100:
            graph_length = 100
        return graph_length

    def print_report(self, collection):
        tags = self.get_categories(collection, self.filter_level)
        if len(tags) == 0:
            print("Filter did not return any matching item")
            return

        max_tag_length = len(max(tags, key=lambda k: len(k[0]))[0]) + 1
        if max_tag_length == 0:
            max_tag_length = 20

        max_duration = next(iter(tags))[1]
        for tag in tags:
            bar_graph = PrintHelpers.get_graph(tag[1], max_duration, self.get_graph_length(max_tag_length))
            line = PrintHelpers.colorize(str(DateTimeUtils.show_timedelta(tag[1])), TerminalColors.GREEN) + ' ' + str(tag[0]).ljust(max_tag_length) \
                + PrintHelpers.colorize(bar_graph, PrintHelpers.get_color())
            print(line)

class DayTimelineReport(TaskReportBase):
    def __init__(self):
        super(DayTimelineReport, self).__init__()
    
    def print_report(self, collection):
        periods = []
        for task in collection.tasks:
            for p in task.periods:
                periods.append([p, task])

        periods = sorted(periods, key=lambda t: t[0].start)

        total_width = os.get_terminal_size().columns
        if total_width < 60:
            print("Terminal is too small to show timeline")
            return
        hour_width = int((total_width - 40) / 11)

        print('Description'.ljust(40) + '8'.ljust(hour_width) + '9'.ljust(hour_width) + '10'.ljust(hour_width) \
            + '11'.ljust(hour_width) + '12'.ljust(hour_width) + '13'.ljust(hour_width) + '14'.ljust(hour_width) \
            + '15'.ljust(hour_width) + '16'.ljust(hour_width) + '17'.ljust(hour_width) + '18'.ljust(hour_width))
        for kv in periods:
            p = kv[0]
            seconds_start = (p.start - datetime.datetime(datetime.datetime.today().year, datetime.datetime.today().month, datetime.datetime.today().day, 8)).seconds
            seconds_end = seconds_start + (p.end - p.start).seconds
            start = int(seconds_start * hour_width / 3600)
            length = int((seconds_end - seconds_start) * hour_width / 3600)
            description = PrintHelpers.max_width(kv[1].description, 40)
            line = PrintHelpers.colorize(description + ' '*start + u'\u2588'*length, PrintHelpers.get_color())
            print(line)

class TagsReport(TaskReportBase):
    def __init__(self):
        super(TagsReport, self).__init__()

    def print_report(self, collection):
        tags = set([])
        for task in collection.tasks:
            for tag in task.tags:
                tags.add(tag)

        for tag in tags:
            print(tag)

    



from datetime import date
from datetime import datetime
from clock_tracking.datetimeutils import DateTimeUtils
from clock_tracking import task

class ClockLogger:
    def __init__(self):
        self.lines = []

    def read_file(self, file):
        try:
            with open(file, 'r') as f:
                self.lines = f.readlines()
            return True
        except Exception:
            return False
        
    def write_file(self, file):
        try:
            with open(file, 'w') as f:
                f.writelines(self.lines)
            return True
        except Exception as e:
            print('Could not write to file ' + str(file) + ': ' + str(e))
            return False

    def date_exists(self):
        today = date.today()
        for l in self.lines:
            if DateTimeUtils.is_date(l) and DateTimeUtils.get_date(l).date() == today:
                return True

    def add_today(self):
        self.lines.append(date.today().strftime('[%Y-%m-%d]') + '\n')

    def edit_current(self, description):
        if len(self.lines) == 0:
            return
        new_description = self.lines[-1][0:5] + ' ' + description + '\n'
        self.lines[-1] = new_description
        print('Edited: ' + self.lines[-1])

    def get_last(self):
        if len(self.lines) == 0:
            return 'Starting'
        
        description = self.lines[-1][6:]
        if '[Stop]' in description and len(self.lines) > 2:
            description = self.lines[-2][6:]

        return description

    # Finds position of entry given the 'at' hour (today)
    def find_position(self, at):
        found_today = False
        at_time = DateTimeUtils.parse_time(date.today(), at)
        index = 0
        for line in self.lines:
            if DateTimeUtils.is_date(line) and DateTimeUtils.get_date(line).date() == date.today():
                found_today = True
            elif found_today and not DateTimeUtils.is_date(line):
                current_time = DateTimeUtils.parse_time(date.today(), line[0:5])
                if current_time > at_time:
                    return index
            
            index = index + 1
        return index

    def add(self, at, description):
        if not self.date_exists():
            self.add_today()

        new_line = at + ' ' + description + '\n'
        self.lines.insert(self.find_position(at), new_line)
        print('Added: ' + new_line)

class ClockReader:
    def __init__(self):
        self.lines = []
    
    def read_file(self, filename):
        try:
            file = open(filename, 'r')
            self.lines = [line for line in file]
        except Exception:
            self.lines = []
            print('Could not open file ' + str(filename))

    def parse(self):
        collection = task.TaskCollection()
        current_day = datetime(1900, 1, 1)
        current_task = None
        for line in self.lines:
            if DateTimeUtils.is_date(line):
                current_task and current_task.discard_last()
                current_day = DateTimeUtils.get_date(line)
                current_task = None
                continue

            time = DateTimeUtils.parse_time(current_day, line[0:5])
            current_task and current_task.finish(time)
            current_task = collection.add_task(task.Task(line[6:].replace('\n', ''), time))

        current_task and current_task.finish(datetime.now())
        return collection




import unittest
from datetime import datetime
from datetime import timedelta
from clock_tracking.filters import *
from clock_tracking.task import Task, Period

class TestDateFilter(unittest.TestCase):
    def test__task_today__returns_true(self):
        filter = DateFilter(datetime.today())
        task = Task('Test', datetime.now())
        self.assertTrue(filter.is_valid(task))

    def test__task_yesterday__returns_false(self):
        filter = DateFilter(datetime.today() - timedelta(days=1))
        task = Task('Test', datetime.now())
        self.assertFalse(filter.is_valid(task))

    def test__task_two_different_periods__returns_task_with_single_period(self):
        filter = DateFilter(datetime.today())
        task = Task('Test', datetime.today() - timedelta(days=1))
        task.finish(datetime.now() - timedelta(days=1))
        task.periods.append(Period(datetime.now()))

        result = filter.get_task(task)
        self.assertEqual(len(result.periods), 1)
        self.assertEqual(result.periods[0].start.date(), datetime.today().date())

class TestPeriodFilter(unittest.TestCase):
    def test__two_periods_within_filter__returns_two_periods(self):
        filter = PeriodFilter(datetime(2022, 1, 1), datetime(2022, 1, 5))
        task = Task('Test', datetime(2022, 1, 2, 10))
        task.finish(datetime(2022, 1, 2, 11))
        task.periods.append(Period(datetime(2022, 1, 3, 10)))
        task.finish(datetime(2022, 1, 3, 11))

        self.assertTrue(filter.is_valid(task))
        new_task = filter.get_task(task)
        self.assertEqual(len(new_task.periods), 2)

    def test__one_period_before_one_in__returns_one_period(self):
        filter = PeriodFilter(datetime(2022, 1, 3), datetime(2022, 1, 5))
        task = Task('Test', datetime(2022, 1, 2, 10))
        task.finish(datetime(2022, 1, 2, 11))
        task.periods.append(Period(datetime(2022, 1, 3, 10)))
        task.finish(datetime(2022, 1, 3, 11))

        self.assertTrue(filter.is_valid(task))
        new_task = filter.get_task(task)
        self.assertEqual(len(new_task.periods), 1)

    def test__two_periods_within_filter_boundaries__returns_two_periods(self):
        filter = PeriodFilter(datetime(2022, 1, 1), datetime(2022, 1, 5))
        task = Task('Test', datetime(2022, 1, 1))
        task.finish(datetime(2022, 1, 1, 11))
        task.periods.append(Period(datetime(2022, 1, 5)))
        task.finish(datetime(2022, 1, 5, 11))

        self.assertTrue(filter.is_valid(task))
        new_task = filter.get_task(task)
        self.assertEqual(len(new_task.periods), 2)
    
class TestIdFilter(unittest.TestCase):
    def test__no_id__returns_true(self):
        filter = IdFilter([])
        task = Task('Test', datetime.today())

        self.assertTrue(filter.is_valid(task))

    def test__valid_id__returns_true(self):
        filter = IdFilter(['.123'])
        task = Task('Test', datetime.today())
        task.ids.append('.123')

        self.assertTrue(filter.is_valid(task))

    def test__multiple_ids_one_valid__returns_true(self):
        filter = IdFilter(['.123'])
        task = Task('Test', datetime.today())
        task.ids.append('.123')
        task.ids.append('.456')

        self.assertTrue(filter.is_valid(task))

    def test__invalid_id__returns_false(self):
        filter = IdFilter(['.123'])
        task = Task('Test', datetime.today())
        task.ids.append('.456')

        self.assertFalse(filter.is_valid(task))

class TestLastFilter(unittest.TestCase):
    def test__one_task__returns_task(self):
        filter = LastFilter(1)
        c = TaskCollection()
        c.tasks.append(Task('Test', datetime.today()))

        c = filter.apply_to(c)
        self.assertTrue(len(c.tasks) == 1)

    def test__two_tasks__returns_one_task(self):
        filter = LastFilter(1)
        c = TaskCollection()
        c.tasks.append(Task('Test', datetime.today()))
        c.tasks.append(Task('Test2', datetime.today()))

        c = filter.apply_to(c)
        self.assertTrue(len(c.tasks) == 1)

    def test__no_task__does_not_throw(self):
        filter = LastFilter(1)
        c = TaskCollection()

        c = filter.apply_to(c)
        self.assertTrue(len(c.tasks) == 0)

    




import unittest
import datetime
from clock_tracking import datetimeutils, options

class TestClockArgumentsGetTargetTime(unittest.TestCase):
    def test__global_target__returns_global_target(self):
        args = options.ClockArguments()
        args.input_args = ['show', '--target', '150:00']
        args.parse()
        target_time = args.get_target_time()

        self.assertFalse(target_time.is_per_day)
        self.assertEqual(target_time.target, datetime.timedelta(0, 150*3600))

    def test__taget_per_day__returns_target_per_day(self):
        args = options.ClockArguments()
        args.input_args = ['show', '--target-per-day', '8:24']
        args.parse()
        target_time = args.get_target_time()

        self.assertTrue(target_time.is_per_day)
        self.assertEqual(target_time.target, datetime.timedelta(0, 8*3600 + 24*60))

class TestClockArgumentsGetFilters(unittest.TestCase):
    def test__today_and_week__returns_today(self):
        args = options.ClockArguments()
        args.input_args = ['show', '--today', '--week']
        args.parse()
        filters = args.get_filters()

        self.assertTrue(len(filters) == 1)
        self.assertEqual(filters[0].date.date(), datetime.datetime.today().date())

    def test__week_and_from__returns_week(self):
        args = options.ClockArguments()
        args.input_args = ['show', '--week', '--from', '2022-01-01']
        args.parse()
        filters = args.get_filters()

        start_week = datetime.datetime.now() - datetime.timedelta(days=datetime.datetime.now().weekday())
        end_week = start_week + datetime.timedelta(days=6)

        self.assertTrue(len(filters) == 1)
        self.assertEqual(filters[0].date_start, start_week.date())
        self.assertEqual(filters[0].date_end, end_week.date())

class TestClockArgumentsParse(unittest.TestCase):
    def test__without_command__adds_default_add(self):
        args = options.ClockArguments()
        args.input_args = ['This', 'is', 'a', 'test']
        args.parse()
        
        self.assertEqual(args.options.command, 'add')

    def test__without_argument__adds_show_command(self):
        args = options.ClockArguments()
        args.input_args = []
        args.parse()
        
        self.assertEqual(args.options.command, 'show')